__license__ = "GNU Affero General Public License http://www.gnu.org/licenses/agpl.html"
__copyright__ = "Copyright (C) 2022 The OctoPrint Project - Released under terms of the AGPLv3 License"

from octoprint.schema import BaseModel
from octoprint.vendor.with_attrs_docs import with_attrs_docs


@with_attrs_docs
class PrintTimeEstimationConfig(BaseModel):
    statsWeighingUntil: float = 0.5
    """Until which percentage to do a weighted mixture of statistical duration (analysis or past prints) with the result from the calculated estimate if that's already available. Utilized to compensate for the fact that the earlier in a print job, the least accuracy even a stable calculated estimate provides."""

    validityRange: float = 0.15
    """Range the assumed percentage (based on current estimated statistical, calculated or mixed total vs elapsed print time so far) needs to be around the actual percentage for the result to be used."""

    forceDumbFromPercent: float = 0.3
    """If no estimate could be calculated until this percentage and no statistical data is available, use dumb linear estimate. Value between 0 and 1.0."""

    forceDumbAfterMin: float = 30.0
    """If no estimate could be calculated until this many minutes into the print and no statistical data is available, use dumb linear estimate."""

    stableThreshold: int = 60
    """Average fluctuation between individual calculated estimates to consider in stable range. Seconds of difference."""


@with_attrs_docs
class EstimationConfig(BaseModel):
    printTime: PrintTimeEstimationConfig = PrintTimeEstimationConfig()
    """Parameters for the print time estimation during an ongoing print job."""



__license__ = "GNU Affero General Public License http://www.gnu.org/licenses/agpl.html"
__copyright__ = "Copyright (C) 2022 The OctoPrint Project - Released under terms of the AGPLv3 License"

from typing import List, Optional

from pydantic import Field

from octoprint.schema import BaseModel
from octoprint.vendor.with_attrs_docs import with_attrs_docs


@with_attrs_docs
class ActionConfig(BaseModel):
    action: str
    """The identifier used internally to identify the action. Set to `divider` to generate a divider in the menu."""

    name: Optional[str] = None
    """The name of the action that will be shown on the menu. Must be set if the action is not a divider."""

    command: Optional[str] = None
    """The command to execute when the action is selected. Must be set if the action is not a divider."""

    async_: bool = Field(False, alias="async")
    """Whether to run the command asynchronously."""

    confirm: Optional[str] = None
    """An optional confirmation message to show before executing the command."""


@with_attrs_docs
class SystemConfig(BaseModel):
    actions: List[ActionConfig] = []
    """A list of system actions to show in the UI."""



__author__ = "Gina Häußge <osd@foosel.net>"
__license__ = "GNU Affero General Public License http://www.gnu.org/licenses/agpl.html"
__copyright__ = "Copyright (C) 2014 The OctoPrint Project - Released under terms of the AGPLv3 License"

import logging

import requests

BRANCH_HEAD_URL = "https://api.github.com/repos/{user}/{repo}/git/refs/heads/{branch}"

logger = logging.getLogger(
    "octoprint.plugins.softwareupdate.version_checks.github_commit"
)


def _get_latest_commit(user, repo, branch, apikey=None):
    from ..exceptions import NetworkError

    headers = {}
    if apikey:
        auth = "token " + apikey
        headers = {"Authorization": auth}

    try:
        r = requests.get(
            BRANCH_HEAD_URL.format(user=user, repo=repo, branch=branch),
            timeout=(3.05, 30),
            headers=headers,
        )
    except requests.ConnectionError as exc:
        raise NetworkError(cause=exc)

    from . import check_github_apiresponse, check_github_ratelimit

    check_github_ratelimit(logger, r)
    check_github_apiresponse(logger, r)

    reference = r.json()
    if "object" not in reference or "sha" not in reference["object"]:
        return None

    return reference["object"]["sha"]


def get_latest(target, check, online=True, credentials=None, *args, **kwargs):
    from ..exceptions import ConfigurationInvalid

    user = check.get("user")
    repo = check.get("repo")

    if user is None or repo is None:
        raise ConfigurationInvalid(
            "Update configuration for {} of type github_commit needs user and repo set and not None".format(
                target
            )
        )

    branch = "master"
    if "branch" in check and check["branch"] is not None:
        branch = check["branch"]

    current = check.get("current")

    information = {
        "local": {
            "name": "Commit {commit}".format(
                commit=current if current is not None else "?"
            ),
            "value": current,
        },
        "remote": {"name": "?", "value": "?"},
        "needs_online": not check.get("offline", False),
    }
    if not online and information["needs_online"]:
        return information, True

    apikey = None
    if credentials:
        apikey = credentials.get("github")

    remote_commit = _get_latest_commit(
        check["user"], check["repo"], branch, apikey=apikey
    )
    remote_name = f"Commit {remote_commit}" if remote_commit is not None else "-"

    information["remote"] = {"name": remote_name, "value": remote_commit}
    is_current = (
        current is not None and current == remote_commit
    ) or remote_commit is None

    logger.debug(f"Target: {target}, local: {current}, remote: {remote_commit}")

    return information, is_current



__author__ = "Gina Häußge <osd@foosel.net>"
__license__ = "GNU Affero General Public License http://www.gnu.org/licenses/agpl.html"
__copyright__ = "Copyright (C) 2015 The OctoPrint Project - Released under terms of the AGPLv3 License"


import logging
import queue
import re
import time
import warnings
from typing import List, Optional, Tuple, Union

import sarge

from octoprint.util import to_bytes, to_unicode
from octoprint.util.platform import CLOSE_FDS

# These regexes are based on the colorama package
# Author: Jonathan Hartley
# License: BSD-3 (https://github.com/tartley/colorama/blob/master/LICENSE.txt)
# Website: https://github.com/tartley/colorama/
_ANSI_CSI_PATTERN = (
    "\001?\033\\[(\\??(?:\\d|;)*)([a-zA-Z])\002?"  # Control Sequence Introducer
)
_ANSI_OSC_PATTERN = "\001?\033\\]((?:.|;)*?)(\x07)\002?"  # Operating System Command
_ANSI_PATTERN = "|".join([_ANSI_CSI_PATTERN, _ANSI_OSC_PATTERN])

_ANSI_REGEX = re.compile(_ANSI_PATTERN)


def clean_ansi(line: Union[str, bytes]) -> Union[str, bytes]:
    """
    Removes ANSI control codes from ``line``.

    Note: This function also still supports an input of ``bytes``, leading to an
    ``output`` of ``bytes``. This if for reasons of backwards compatibility only,
    should no longer be used and considered to be deprecated and to be removed in
    a future version of OctoPrint. A warning will be logged.

    Parameters:
        line (str or bytes): the line to process

    Returns:
        (str or bytes) The line without any ANSI control codes

    .. versionchanged:: 1.8.0

       Usage as ``clean_ansi(line: bytes) -> bytes`` is now deprecated and will be removed
       in a future version of OctoPrint.
    """
    # TODO: bytes support is deprecated, remove in 2.0.0
    if isinstance(line, bytes):
        warnings.warn(
            "Calling clean_ansi with bytes is deprecated, call with str instead",
            DeprecationWarning,
            stacklevel=2,
        )
        return to_bytes(_ANSI_REGEX.sub("", to_unicode(line)))
    return _ANSI_REGEX.sub("", line)


class CommandlineError(Exception):
    """
    Raised by :py:func:`~octoprint.util.commandline.CommandLineCaller.checked_call` on non zero return codes

    Arguments:
        returncode (int): the return code of the command
        stdout (str): the stdout output produced by the command
        stderr (str): the stderr output produced by the command
    """

    def __init__(self, returncode, stdout, stderr):
        self.returncode = returncode
        self.stdout = stdout
        self.stderr = stderr


class CommandlineCaller:
    """
    The CommandlineCaller is a utility class that allows running command line commands while logging their stdout
    and stderr via configurable callback functions.

    Callbacks are expected to have a signature matching

    .. code-block:: python

       def callback(*lines):
           do_something_with_the_passed_lines()

    The class utilizes sarge underneath.

    Example:

    .. code-block:: python

       from octoprint.util.commandline import CommandLineCaller, CommandLineError

       def log(prefix, *lines):
           for line in lines:
               print("{} {}".format(prefix, line))

       def log_stdout(*lines):
           log(">>>", *lines)

       def log_stderr(*lines):
           log("!!!", *lines)

       def log_call(*lines)
           log("---", *lines)

       caller = CommandLineCaller()
       caller.on_log_call = log_call
       caller.on_log_stdout = log_stdout
       caller.on_log_stderr = log_stderr

       try:
           caller.checked_call(["some", "command", "with", "parameters"])
       except CommandLineError as err:
           print("Command returned {}".format(err.returncode))
       else:
           print("Command finished successfully")
    """

    def __init__(self):
        self._logger = logging.getLogger(__name__)

        self.on_log_call = lambda *args, **kwargs: None
        """Callback for the called command line"""

        self.on_log_stdout = lambda *args, **kwargs: None
        """Callback for stdout output"""

        self.on_log_stderr = lambda *args, **kwargs: None
        """Callback for stderr output"""

    def checked_call(
        self, command: Union[str, List[str], Tuple[str]], **kwargs
    ) -> Tuple[int, List[str], List[str]]:
        """
        Calls a command and raises an error if it doesn't return with return code 0

        Args:
            command (list, tuple or str): command to call
            kwargs (dict): additional keyword arguments to pass to the sarge ``run`` call (note that ``_async``,
                           ``stdout`` and ``stderr`` will be overwritten)

        Returns:
            (tuple) a 3-tuple of return code, full stdout and full stderr output

        Raises:
            CommandlineError
        """
        returncode, stdout, stderr = self.call(command, **kwargs)

        if returncode != 0:
            raise CommandlineError(returncode, stdout, stderr)

        return returncode, stdout, stderr

    def call(
        self,
        command: Union[str, List[str], Tuple[str]],
        delimiter: bytes = b"\n",
        buffer_size: int = -1,
        logged: bool = True,
        **kwargs,
    ) -> Tuple[Optional[int], List[str], List[str]]:
        """
        Calls a command

        Args:
            command (list, tuple or str): command to call
            kwargs (dict): additional keyword arguments to pass to the sarge ``run`` call (note that ``_async``,
                           ``stdout`` and ``stderr`` will be overwritten)

        Returns:
            (tuple) a 3-tuple of return code, full stdout and full stderr output
        """

        p = self.non_blocking_call(
            command, delimiter=delimiter, buffer_size=buffer_size, logged=logged, **kwargs
        )
        if p is None:
            return None, [], []

        all_stdout = []
        all_stderr = []

        def process_lines(lines, logger):
            if not lines:
                return []
            processed = self._preprocess_lines(
                *map(lambda x: to_unicode(x, errors="replace"), lines)
            )
            if logged:
                logger(*processed)
            return list(processed)

        def process_stdout(lines):
            return process_lines(lines, self._log_stdout)

        def process_stderr(lines):
            return process_lines(lines, self._log_stderr)

        try:
            while p.returncode is None:
                all_stderr += process_stderr(p.stderr.readlines(timeout=0.5))
                all_stdout += process_stdout(p.stdout.readlines(timeout=0.5))
                p.commands[0].poll()

        finally:
            p.close()

        all_stderr += process_stderr(p.stderr.readlines())
        all_stdout += process_stdout(p.stdout.readlines())

        return p.returncode, all_stdout, all_stderr

    def non_blocking_call(
        self,
        command: Union[str, List, Tuple],
        delimiter: bytes = b"\n",
        buffer_size: int = -1,
        logged: bool = True,
        **kwargs,
    ) -> Optional[sarge.Pipeline]:
        if isinstance(command, (list, tuple)):
            joined_command = " ".join(command)
        else:
            joined_command = command
        self._logger.debug(f"Calling: {joined_command}")

        if logged:
            self.on_log_call(joined_command)

        kwargs.update(
            {
                "close_fds": CLOSE_FDS,
                "async_": True,
                "stdout": DelimiterCapture(delimiter=delimiter, buffer_size=buffer_size),
                "stderr": DelimiterCapture(delimiter=delimiter, buffer_size=buffer_size),
            }
        )

        p = sarge.run(command, **kwargs)
        while len(p.commands) == 0:
            # somewhat ugly... we can't use wait_events because
            # the events might not be all set if an exception
            # by sarge is triggered within the async process
            # thread
            time.sleep(0.01)

        # by now we should have a command, let's wait for its
        # process to have been prepared
        p.commands[0].process_ready.wait()

        if not p.commands[0].process:
            # the process might have been set to None in case of any exception
            self._logger.error(f"Error while trying to run command {joined_command}")
            return None

        return p

    def _log_stdout(self, *lines):
        self.on_log_stdout(*lines)

    def _log_stderr(self, *lines):
        self.on_log_stderr(*lines)

    def _preprocess_lines(self, *lines):
        return lines


class DelimiterCapture(sarge.Capture):
    def __init__(self, delimiter=b"\n", *args, **kwargs):
        self._delimiter = delimiter
        sarge.Capture.__init__(self, *args, **kwargs)

    def readline(self, size=-1, block=True, timeout=None):
        if not self.streams_open():
            block = False
            timeout = None
        else:
            timeout = timeout or self.timeout
        if self.current is None:
            try:
                self.current = self.buffer.get(block, timeout)
            except queue.Empty:
                self.current = b""
        while self._delimiter not in self.current:
            try:
                self.current += self.buffer.get(block, timeout)
            except queue.Empty:
                break
        if self._delimiter not in self.current:
            result = self.current
            self.current = None
        else:
            i = self.current.index(self._delimiter)
            if 0 < size < i:
                i = size - 1
            result = self.current[: i + 1]
            self.current = self.current[i + 1 :]
        return result



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals

"""
    sockjs.tornado.transports.xhr
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    Xhr-Polling transport implementation
"""
import logging

from octoprint.vendor.sockjs.tornado import proto
from octoprint.vendor.sockjs.tornado.transports import pollingbase
from octoprint.vendor.sockjs.tornado.util import bytes_to_str, no_auto_finish

LOG = logging.getLogger("tornado.general")

class XhrPollingTransport(pollingbase.PollingTransportBase):
    """xhr-polling transport implementation"""
    name = 'xhr'

    @no_auto_finish
    def post(self, session_id):
        # Start response
        self.preflight()
        self.handle_session_cookie()
        self.disable_cache()

        # Get or create session without starting heartbeat
        if not self._attach_session(session_id, False):
            return

        # Might get already detached because connection was closed in on_open
        if not self.session:
            return

        if not self.session.send_queue:
            self.session.start_heartbeat()
        else:
            self.session.flush()

    def send_pack(self, message, binary=False):
        if binary:
            raise Exception('binary not supported for XhrPollingTransport')

        self.active = False

        try:
            self.set_header('Content-Type', 'application/javascript; charset=UTF-8')
            self.set_header('Content-Length', len(message) + 1)
            self.write(message + '\n')
            self.flush().add_done_callback(self.send_complete)
        except IOError:
            # If connection dropped, make sure we close offending session instead
            # of propagating error all way up.
            self.session.delayed_close()


class XhrSendHandler(pollingbase.PollingTransportBase):
    def post(self, session_id):
        self.preflight()
        self.handle_session_cookie()
        self.disable_cache()

        session = self._get_session(session_id)

        if session is None or session.is_closed:
            self.set_status(404)
            return

        data = self.request.body
        if not data:
            self.write("Payload expected.")
            self.set_status(500)
            return

        try:
            messages = proto.json_decode(bytes_to_str(data))
        except Exception:
            # TODO: Proper error handling
            self.write("Broken JSON encoding.")
            self.set_status(500)
            return

        try:
            session.on_messages(messages)
        except Exception:
            LOG.exception('XHR incoming')
            session.close()

            self.set_status(500)
            return

        self.set_status(204)
        self.set_header('Content-Type', 'text/plain; charset=UTF-8')



__author__ = "Gina Häußge <osd@foosel.net>"
__license__ = "GNU Affero General Public License http://www.gnu.org/licenses/agpl.html"
__copyright__ = "Copyright (C) 2014 The OctoPrint Project - Released under terms of the AGPLv3 License"


from octoprint.settings import settings


class PrintTimeEstimator:
    """
    Estimator implementation.

    Subclass this and register via the octoprint.printer.estimation.factory hook to provide your own implementation.
    """

    def __init__(self, job_type):
        self.stats_weighing_until = settings().getFloat(
            ["estimation", "printTime", "statsWeighingUntil"]
        )
        self.validity_range = settings().getFloat(
            ["estimation", "printTime", "validityRange"]
        )
        self.force_dumb_from_percent = settings().getFloat(
            ["estimation", "printTime", "forceDumbFromPercent"]
        )
        self.force_dumb_after_min = settings().getFloat(
            ["estimation", "printTime", "forceDumbAfterMin"]
        )

        threshold = None
        rolling_window = None
        countdown = None

        if job_type == "local" or job_type == "sdcard":
            # we are happy if the average of the estimates stays within 60s of the prior one
            threshold = settings().getFloat(
                ["estimation", "printTime", "stableThreshold"]
            )

            if job_type == "sdcard":
                # we are interesting in a rolling window of roughly the last 15s, so the number of entries has to be derived
                # by that divided by the sd status polling interval
                interval = settings().getFloat(["serial", "timeout", "sdStatus"])
                if interval <= 0:
                    interval = 1.0
                rolling_window = int(15 // interval)
                if rolling_window < 1:
                    rolling_window = 1

                # we are happy when one rolling window has been stable
                countdown = rolling_window

        self._data = TimeEstimationHelper(
            rolling_window=rolling_window, countdown=countdown, threshold=threshold
        )

    def estimate(
        self,
        progress,
        printTime,
        cleanedPrintTime,
        statisticalTotalPrintTime,
        statisticalTotalPrintTimeType,
    ):
        """
        Tries to estimate the print time left for the print job

        This is somewhat horrible since accurate print time estimation is pretty much impossible to
        achieve, considering that we basically have only two data points (current progress in file and
        time needed for that so far - former prints or a file analysis might not have happened or simply
        be completely impossible e.g. if the file is stored on the printer's SD card) and
        hence can only do a linear estimation of a completely non-linear process. That's a recipe
        for inaccurate predictions right there. Yay.

        Anyhow, here's how this implementation works. This method gets the current progress in the
        printed file (percentage based on bytes read vs total bytes), the print time that elapsed,
        the same print time with the heat up times subtracted (if possible) and if available also
        some statistical total print time (former prints or a result from the GCODE analysis).

          1. First get an "intelligent" estimate based on the :class:`~octoprint.printer.estimation.TimeEstimationHelper`.
             That thing tries to detect if the estimation based on our progress and time needed for that becomes
             stable over time through a rolling window and only returns a result once that appears to be the
             case.
          2. If we have any statistical data (former prints or a result from the GCODE analysis)
             but no intelligent estimate yet, we'll use that for the next step. Otherwise, up to a certain percentage
             in the print we do a percentage based weighing of the statistical data and the intelligent
             estimate - the closer to the beginning of the print, the more precedence for the statistical
             data, the closer to the cut off point, the more precedence for the intelligent estimate. This
             is our preliminary total print time.
          3. If the total print time is set, we do a sanity check for it. Based on the total print time
             estimate and the time we already spent printing, we calculate at what percentage we SHOULD be
             and compare that to the percentage at which we actually ARE. If it's too far off, our total
             can't be trusted and we fall back on the dumb estimate. Same if the time we spent printing is
             already higher than our total estimate.
          4. If we do NOT have a total print time estimate yet but we've been printing for longer than
             a configured amount of minutes or are further in the file than a configured percentage, we
             also use the dumb estimate for now.

        Yes, all this still produces horribly inaccurate results. But we have to do this live during the print and
        hence can't produce to much computational overhead, we do not have any insight into the firmware implementation
        with regards to planner setup and acceleration settings, we might not even have access to the printed file's
        contents and such we need to find something that works "mostly" all of the time without costing too many
        resources. Feel free to propose a better solution within the above limitations (and I mean that, this solution
        here makes me unhappy).

        Args:
            progress (float or None): Current percentage in the printed file
            printTime (float or None): Print time elapsed so far
            cleanedPrintTime (float or None): Print time elapsed minus the time needed for getting up to temperature
                (if detectable).
            statisticalTotalPrintTime (float or None): Total print time of past prints against same printer profile,
                or estimated total print time from GCODE analysis.
            statisticalTotalPrintTimeType (str or None): Type of statistical print time, either "average" (total time
                of former prints) or "analysis"

        Returns:
            (2-tuple) estimated print time left or None if not proper estimate could be made at all, origin of estimation
        """

        if (
            progress is None
            or progress == 0
            or printTime is None
            or cleanedPrintTime is None
        ):
            return None, None

        dumbTotalPrintTime = printTime / progress
        estimatedTotalPrintTime = self.estimate_total(progress, cleanedPrintTime)
        totalPrintTime = estimatedTotalPrintTime

        printTimeLeftOrigin = "estimate"
        if statisticalTotalPrintTime is not None:
            if estimatedTotalPrintTime is None:
                # no estimate yet, we'll use the statistical total
                totalPrintTime = statisticalTotalPrintTime
                printTimeLeftOrigin = statisticalTotalPrintTimeType

            else:
                if progress < self.stats_weighing_until:
                    # still inside weighing range, use part stats, part current estimate
                    sub_progress = progress * (1 / self.stats_weighing_until)
                    if sub_progress > 1.0:
                        sub_progress = 1.0
                    printTimeLeftOrigin = "mixed-" + statisticalTotalPrintTimeType
                else:
                    # use only the current estimate
                    sub_progress = 1.0
                    printTimeLeftOrigin = "estimate"

                # combine
                totalPrintTime = (
                    -sub_progress * statisticalTotalPrintTime
                    + sub_progress * estimatedTotalPrintTime
                )

        printTimeLeft = None
        if totalPrintTime is not None:
            # sanity check current total print time estimate
            assumed_progress = cleanedPrintTime / totalPrintTime
            min_progress = progress - self.validity_range
            max_progress = progress + self.validity_range

            if (
                min_progress <= assumed_progress <= max_progress
                and totalPrintTime > cleanedPrintTime
            ):
                # appears sane, we'll use it
                printTimeLeft = totalPrintTime - cleanedPrintTime

            else:
                # too far from the actual progress or negative,
                # we use the dumb print time instead
                printTimeLeft = dumbTotalPrintTime - cleanedPrintTime
                printTimeLeftOrigin = "linear"

        else:
            printTimeLeftOrigin = "linear"
            if (
                progress > self.force_dumb_from_percent
                or cleanedPrintTime >= self.force_dumb_after_min * 60
            ):
                # more than x% or y min printed and still no real estimate, ok, we'll use the dumb variant :/
                printTimeLeft = dumbTotalPrintTime - cleanedPrintTime

        if printTimeLeft is not None and printTimeLeft < 0:
            # shouldn't actually happen, but let's make sure
            printTimeLeft = None

        return printTimeLeft, printTimeLeftOrigin

    def estimate_total(self, progress, printTime):
        if not progress or not printTime or not self._data:
            return None
        else:
            return self._data.update(printTime / progress)


class TimeEstimationHelper:
    STABLE_THRESHOLD = 0.1
    STABLE_COUNTDOWN = 250
    STABLE_ROLLING_WINDOW = 250

    def __init__(self, rolling_window=None, countdown=None, threshold=None):
        if rolling_window is None:
            rolling_window = self.__class__.STABLE_ROLLING_WINDOW
        if countdown is None:
            countdown = self.__class__.STABLE_COUNTDOWN
        if threshold is None:
            threshold = self.__class__.STABLE_THRESHOLD

        self._rolling_window = rolling_window
        self._countdown = countdown
        self._threshold = threshold

        import collections

        self._distances = collections.deque([], self._rolling_window)
        self._totals = collections.deque([], self._rolling_window)
        self._sum_total = 0
        self._count = 0
        self._stable_counter = -1

    def is_stable(self):
        return self._stable_counter >= self._countdown

    def update(self, new_estimate):
        old_average_total = self.average_total

        self._sum_total += new_estimate
        self._totals.append(new_estimate)
        self._count += 1

        if old_average_total:
            self._distances.append(abs(self.average_total - old_average_total))

        if -self._threshold < self.average_distance < self._threshold:
            self._stable_counter += 1
        else:
            self._stable_counter = -1

        if self.is_stable():
            return self.average_total_rolling
        else:
            return None

    @property
    def average_total(self):
        if not self._count:
            return 0
        else:
            return self._sum_total / self._count

    @property
    def average_total_rolling(self):
        if not self._count or self._count < self._rolling_window or not len(self._totals):
            return -1
        else:
            return sum(self._totals) / len(self._totals)

    @property
    def average_distance(self):
        if (
            not self._count
            or self._count < self._rolling_window + 1
            or not len(self._distances)
        ):
            return -1
        else:
            return sum(self._distances) / len(self._distances)



__license__ = "GNU Affero General Public License http://www.gnu.org/licenses/agpl.html"
__copyright__ = "Copyright (C) 2022 The OctoPrint Project - Released under terms of the AGPLv3 License"

from typing import Dict, Optional

from octoprint.schema import BaseModel
from octoprint.vendor.with_attrs_docs import with_attrs_docs


@with_attrs_docs
class GcodeScriptsConfig(BaseModel):
    afterPrinterConnected: Optional[str] = None
    beforePrinterDisconnected: Optional[str] = None
    beforePrintStarted: Optional[str] = None
    afterPrintCancelled: Optional[
        str
    ] = "; disable motors\nM84\n\n;disable all heaters\n{% snippet 'disable_hotends' %}\n{% snippet 'disable_bed' %}\n;disable fan\nM106 S0"
    afterPrintDone: Optional[str] = None
    beforePrintPaused: Optional[str] = None
    afterPrintResumed: Optional[str] = None
    beforeToolChange: Optional[str] = None
    afterToolChange: Optional[str] = None
    snippets: Dict[str, str] = {
        "disable_hotends": "{% if printer_profile.extruder.sharedNozzle %}M104 T0 S0\n{% else %}{% for tool in range(printer_profile.extruder.count) %}M104 T{{ tool }} S0\n{% endfor %}{% endif %}",
        "disable_bed": "{% if printer_profile.heatedBed %}M140 S0\n{% endif %}",
    }


@with_attrs_docs
class ScriptsConfig(BaseModel):
    gcode: GcodeScriptsConfig = GcodeScriptsConfig()



__license__ = "GNU Affero General Public License http://www.gnu.org/licenses/agpl.html"
__copyright__ = "Copyright (C) 2015 The OctoPrint Project - Released under terms of the AGPLv3 License"

import collections
import logging
import re
import threading

import psutil
from flask import abort, jsonify, request, url_for
from flask_babel import gettext

from octoprint.access.permissions import Permissions
from octoprint.logging import prefix_multilines
from octoprint.plugin import plugin_manager
from octoprint.server import NO_CONTENT
from octoprint.server.api import api
from octoprint.server.util.flask import get_remote_address, no_firstrun_access
from octoprint.settings import settings as s
from octoprint.systemcommands import system_command_manager
from octoprint.util.commandline import CommandlineCaller


@api.route("/system/usage", methods=["GET"])
@no_firstrun_access
@Permissions.SYSTEM.require(403)
def readUsageForFolders():
    return jsonify(usage=_usageForFolders())


@api.route("/system/info", methods=["GET"])
@no_firstrun_access
@Permissions.SYSTEM.require(403)
def getSystemInfo():
    from octoprint.cli.systeminfo import get_systeminfo
    from octoprint.server import (
        connectivityChecker,
        environmentDetector,
        printer,
        safe_mode,
    )
    from octoprint.util import dict_flatten

    systeminfo = get_systeminfo(
        environmentDetector,
        connectivityChecker,
        s(),
        {
            "browser.user_agent": request.headers.get("User-Agent"),
            "octoprint.safe_mode": safe_mode is not None,
            "systeminfo.generator": "systemapi",
        },
    )

    if printer and printer.is_operational():
        firmware_info = printer.firmware_info
        if firmware_info:
            systeminfo.update(
                dict_flatten({"firmware": firmware_info["name"]}, prefix="printer")
            )

    return jsonify(systeminfo=systeminfo)


def _usageForFolders():
    data = {}
    for folder_name in s().get(["folder"]).keys():
        path = s().getBaseFolder(folder_name, check_writable=False)
        if path is not None:
            usage = psutil.disk_usage(path)
            data[folder_name] = {"free": usage.free, "total": usage.total}
    return data


@api.route("/system", methods=["POST"])
@no_firstrun_access
@Permissions.SYSTEM.require(403)
def performSystemAction():
    logging.getLogger(__name__).warning(
        "Deprecated API call to /api/system made by {}, should be migrated to use /system/commands/custom/<action>".format(
            get_remote_address(request)
        )
    )

    data = request.get_json(silent=True)
    if data is None:
        data = request.values

    if "action" not in data:
        abort(400, description="action is missing")

    return executeSystemCommand("custom", data["action"])


@api.route("/system/commands", methods=["GET"])
@no_firstrun_access
@Permissions.SYSTEM.require(403)
def retrieveSystemCommands():
    return jsonify(
        core=_to_client_specs(_get_core_command_specs()),
        plugin=_to_client_specs(_get_plugin_command_specs()),
        custom=_to_client_specs(_get_custom_command_specs()),
    )


@api.route("/system/commands/<string:source>", methods=["GET"])
@no_firstrun_access
@Permissions.SYSTEM.require(403)
def retrieveSystemCommandsForSource(source):
    if source == "core":
        specs = _get_core_command_specs()
    elif source == "custom":
        specs = _get_custom_command_specs()
    elif source == "plugin":
        specs = _get_plugin_command_specs()
    else:
        abort(404)

    return jsonify(_to_client_specs(specs))


@api.route("/system/commands/<string:source>/<string:command>", methods=["POST"])
@no_firstrun_access
@Permissions.SYSTEM.require(403)
def executeSystemCommand(source, command):
    logger = logging.getLogger(__name__)

    if command == "divider":
        abort(400, description="Dividers cannot be executed")

    command_spec = _get_command_spec(source, command)
    if not command_spec:
        abort(404)

    if "command" not in command_spec:
        abort(
            500, description="Command does not define a command to execute, can't proceed"
        )

    do_async = command_spec.get("async", False)
    do_ignore = command_spec.get("ignore", False)
    debug = command_spec.get("debug", False)

    if logger.isEnabledFor(logging.DEBUG) or debug:
        logger.info(
            "Performing command for {}:{}: {}".format(
                source, command, command_spec["command"]
            )
        )
    else:
        logger.info(f"Performing command for {source}:{command}")

    try:
        if "before" in command_spec and callable(command_spec["before"]):
            command_spec["before"]()
    except Exception as e:
        if not do_ignore:
            error = f'Command "before" for {source}:{command} failed: {e}'
            logger.warning(error)
            abort(500, description=error)

    try:

        def execute():
            # we run this with shell=True since we have to trust whatever
            # our admin configured as command and since we want to allow
            # shell-alike handling here...
            return_code, stdout_lines, stderr_lines = CommandlineCaller().call(
                command_spec["command"], shell=True
            )

            if not do_ignore and return_code != 0:
                stdout = "\n".join(stdout_lines)
                stderr = "\n".join(stderr_lines)
                error = f"Command for {source}:{command} failed with return code {return_code}:\n\nSTDOUT:\n{stdout}\n\nSTDERR:\n{stderr}"
                logger.warning(prefix_multilines(error, prefix="! "))
                if not do_async:
                    raise CommandFailed(error)

        if do_async:
            thread = threading.Thread(target=execute)
            thread.daemon = True
            thread.start()

        else:
            try:
                execute()
            except CommandFailed as exc:
                abort(500, exc.error)

    except Exception as e:
        if not do_ignore:
            error = f"Command for {source}:{command} failed: {e}"
            logger.warning(error)
            abort(500, error)

    return NO_CONTENT


def _to_client_specs(specs):
    result = list()
    for spec in specs.values():
        if "action" not in spec or "source" not in spec:
            continue
        copied = {
            k: v for k, v in spec.items() if k in ("source", "action", "name", "confirm")
        }
        copied["resource"] = url_for(
            ".executeSystemCommand",
            source=spec["source"],
            command=spec["action"],
            _external=True,
        )
        result.append(copied)
    return result


def _get_command_spec(source, action):
    if source == "core":
        return _get_core_command_spec(action)
    elif source == "custom":
        return _get_custom_command_spec(action)
    elif source == "plugin":
        return _get_plugin_command_spec(action)
    else:
        return None


def _get_core_command_specs():
    def enable_safe_mode():
        s().set(["server", "startOnceInSafeMode"], True)
        s().save()

    commands = collections.OrderedDict(
        shutdown={
            "command": system_command_manager().get_system_shutdown_command(),
            "name": gettext("Shutdown system"),
            "confirm": gettext(
                "<strong>You are about to shutdown the system.</strong></p><p>This action may disrupt any ongoing print jobs (depending on your printer's controller and general setup that might also apply to prints run directly from your printer's internal storage)."
            ),
        },
        reboot={
            "command": system_command_manager().get_system_restart_command(),
            "name": gettext("Reboot system"),
            "confirm": gettext(
                "<strong>You are about to reboot the system.</strong></p><p>This action may disrupt any ongoing print jobs (depending on your printer's controller and general setup that might also apply to prints run directly from your printer's internal storage)."
            ),
        },
        restart={
            "command": system_command_manager().get_server_restart_command(),
            "name": gettext("Restart OctoPrint"),
            "confirm": gettext(
                "<strong>You are about to restart the OctoPrint server.</strong></p><p>This action may disrupt any ongoing print jobs (depending on your printer's controller and general setup that might also apply to prints run directly from your printer's internal storage)."
            ),
        },
        restart_safe={
            "command": system_command_manager().get_server_restart_command(),
            "name": gettext("Restart OctoPrint in safe mode"),
            "confirm": gettext(
                "<strong>You are about to restart the OctoPrint server in safe mode.</strong></p><p>This action may disrupt any ongoing print jobs (depending on your printer's controller and general setup that might also apply to prints run directly from your printer's internal storage)."
            ),
            "before": enable_safe_mode,
        },
    )

    available_commands = collections.OrderedDict()
    for action, spec in commands.items():
        if not spec["command"]:
            continue
        spec.update({"action": action, "source": "core", "async": True, "debug": True})
        available_commands[action] = spec
    return available_commands


def _get_core_command_spec(action):
    available_actions = _get_core_command_specs()
    if action not in available_actions:
        logging.getLogger(__name__).warning(
            "Command for core action {} is not configured, you need to configure the command before it can be used".format(
                action
            )
        )
        return None

    return available_actions[action]


_plugin_action_regex = re.compile(r"[a-z0-9_-]+")


def _get_plugin_command_specs(plugin=None):
    specs = collections.OrderedDict()

    hooks = plugin_manager().get_hooks("octoprint.system.additional_commands")
    if plugin is not None:
        if plugin in hooks:
            hooks = {plugin: hooks[plugin]}
        else:
            hooks = {}

    for name, hook in hooks.items():
        try:
            plugin_specs = hook()
            for spec in plugin_specs:
                action = spec.get("action")
                if (
                    not action
                    or action == "divider"
                    or not _plugin_action_regex.match(action)
                ):
                    continue
                action = name.lower() + ":" + action

                copied = dict(spec)
                copied["source"] = "plugin"
                copied["action"] = action
                specs[action] = copied
        except Exception:
            logging.getLogger(__name__).exception(
                f"Error while fetching additional actions from plugin {name}",
                extra={"plugin": name},
            )
    return specs


def _get_plugin_command_spec(action):
    plugin = action.split(":", 1)[0]
    available_actions = _get_plugin_command_specs(plugin=plugin)
    return available_actions.get(action)


def _get_custom_command_specs():
    specs = collections.OrderedDict()
    dividers = 0
    for spec in s().get(["system", "actions"]):
        if "action" not in spec:
            continue
        copied = dict(spec)
        copied["source"] = "custom"

        action = spec["action"]
        if action == "divider":
            dividers += 1
            action = f"divider_{dividers}"
        specs[action] = copied
    return specs


def _get_custom_command_spec(action):
    available_actions = _get_custom_command_specs()
    return available_actions.get(action)


class CommandFailed(Exception):
    def __init__(self, error):
        self.error = error



__license__ = "GNU Affero General Public License http://www.gnu.org/licenses/agpl.html"
__copyright__ = "Copyright (C) 2022 The OctoPrint Project - Released under terms of the AGPLv3 License"

from typing import List, Optional

from octoprint.schema import BaseModel
from octoprint.vendor.with_attrs_docs import with_attrs_docs


@with_attrs_docs
class AccessControlConfig(BaseModel):
    salt: Optional[str] = None
    """Secret salt used for password hashing. **DO NOT TOUCH!** If changed you will no longer be able to log in with your existing accounts. Default unset, generated on first run."""

    userManager: str = "octoprint.access.users.FilebasedUserManager"
    """The user manager implementation to use for accessing user information. Currently only a filebased user manager is implemented which stores configured accounts in a YAML file (Default: `users.yaml` in the default configuration folder)."""

    groupManager: str = "octoprint.access.groups.FilebasedGroupManager"
    """The group manager implementation to use for accessing group information. Currently only a filebased user manager is implemented which stores configured groups in a YAML file (Default: `groups.yaml` in the default configuration folder)."""

    permissionManager: str = "octoprint.access.permissions.PermissionManager"
    """The permission manager implementation to use."""

    userfile: Optional[str] = None
    """The YAML user file to use. If left out defaults to `users.yaml` in the default configuration folder."""

    groupfile: Optional[str] = None
    """The YAML group file to use. If left out defaults to `groups.yaml` in the default configuration folder."""

    autologinLocal: bool = False
    """If set to true, will automatically log on clients originating from any of the networks defined in `localNetworks` as the user defined in `autologinAs`."""

    localNetworks: List[str] = ["127.0.0.0/8", "::1/128"]
    """A list of networks or IPs for which an automatic logon as the user defined in `autologinAs` will take place. If available OctoPrint will evaluate the `X-Forwarded-For` HTTP header for determining the client's IP address. Defaults to anything originating from localhost."""

    autologinAs: Optional[str] = None
    """The name of the user to automatically log on clients originating from `localNetworks` as. Must be the name of one of your configured users."""

    trustBasicAuthentication: bool = False
    """Whether to trust Basic Authentication headers. If you have setup Basic Authentication in front of OctoPrint and the user names you use there match OctoPrint accounts, by setting this to true users will be logged into OctoPrint as the user user during Basic Authentication. **ONLY ENABLE THIS** if your OctoPrint instance is only accessible through a connection locked down through Basic Authentication!"""

    checkBasicAuthenticationPassword: bool = True
    """Whether to also check the password provided through Basic Authentication, if the Basic Authentication header is to be trusted. Disabling this will only match the user name in the Basic Authentication header and login the user without further checks, thus disable with caution."""

    trustRemoteUser: bool = False
    """Whether to trust remote user headers. If you have setup authentication in front of OctoPrint and the user names you use there match OctoPrint accounts, by setting this to true users will be logged into OctoPrint as the user provided in the header. **ONLY ENABLE THIS** if your OctoPrint instance is only accessible through a connection locked down through an authenticating reverse proxy!"""

    remoteUserHeader: str = "REMOTE_USER"
    """Header used by the reverse proxy to convey the authenticated user."""

    addRemoteUsers: bool = False
    """If a remote user is not found, add them. Use this only if all users from the remote system can use OctoPrint."""
